Attribute VB_Name = "Module1"
Option Explicit

Public Const PI As Single = 3.14159

Public N As Long
Public Iter As Long
Public diff As Single
Public visc As Single
Public dt As Single

Public s() As Single
Public density() As Single
Public Vx() As Single
Public Vy() As Single
Public Vx0() As Single
Public Vy0() As Single

Public Function IX(x As Long, y As Long) As Long

IX = x + y * N

End Function

Public Sub Init()

Randomize Timer

N = 180

Form1.Picture1.Width = N * Screen.TwipsPerPixelX
Form1.Picture1.Height = N * Screen.TwipsPerPixelY

Call InitPB(Form1.Picture1, Form1.Picture1.ScaleWidth, Form1.Picture1.ScaleHeight)

Iter = 4
dt = 0.2
diff = 0.000001
visc = 0.0000001

ReDim s(N * N)
ReDim density(N * N)
ReDim Vx(N * N)
ReDim Vy(N * N)
ReDim Vx0(N * N)
ReDim Vy0(N * N)

End Sub

Public Sub step()

Call diffuse(1, Vx0, Vx, visc, dt)
Call diffuse(2, Vy0, Vy, visc, dt)

Call project(Vx0, Vy0, Vx, Vy)

Call advect(1, Vx, Vx0, Vx0, Vy0, dt)
Call advect(2, Vy, Vy0, Vx0, Vy0, dt)

Call project(Vx, Vy, Vx0, Vy0)
Call diffuse(0, s, density, diff, dt)
Call advect(0, density, s, Vx, Vy, dt)

End Sub

Public Sub addDensity(x As Long, y As Long, amount As Single)

Dim Index As Long

Index = IX(x, y)
    
density(Index) = density(Index) + amount

End Sub

Public Sub addVelocity(x As Long, y As Long, amountX As Single, amountY As Single)

Dim Index As Long

Index = IX(x, y)

Vx(Index) = Vx(Index) + amountX
Vy(Index) = Vy(Index) + amountY

End Sub

Public Sub diffuse(b As Long, x() As Single, x0() As Single, diff As Single, dt As Single)

Dim a As Single

a = dt * diff * (N - 2) * (N - 2)
Call lin_solve(b, x, x0, a, 1 + 6 * a)
  
End Sub

Public Sub lin_solve(b As Long, x() As Single, x0() As Single, a As Single, c As Single)

Dim cRecip As Single
Dim t As Long
Dim j As Long
Dim i As Long

cRecip = 1 / c

For t = 0 To Iter - 1
    For j = 1 To N - 2
        For i = 1 To N - 2
            x(IX(i, j)) = (x0(IX(i, j)) + a * (x(IX(i + 1, j)) + x(IX(i - 1, j)) + x(IX(i, j + 1)) + x(IX(i, j - 1)))) * cRecip
        Next
    Next

    'Call set_bnd(b, x)
Next

End Sub

Public Sub project(velocX() As Single, velocY() As Single, p() As Single, div() As Single)

Dim j As Long
Dim i As Long

For j = 1 To N - 2
    For i = 1 To N - 2
        div(IX(i, j)) = (-0.5 * (velocX(IX(i + 1, j)) - velocX(IX(i - 1, j)) + velocY(IX(i, j + 1)) - velocY(IX(i, j - 1)))) / N
        p(IX(i, j)) = 0
    Next
Next

'Call set_bnd(0, div)
'Call set_bnd(0, p)
Call lin_solve(0, p, div, 1, 6)

For j = 1 To N - 2
    For i = 1 To N - 2
        velocX(IX(i, j)) = velocX(IX(i, j)) - 0.5 * (p(IX(i + 1, j)) - p(IX(i - 1, j))) * N
        velocY(IX(i, j)) = velocY(IX(i, j)) - 0.5 * (p(IX(i, j + 1)) - p(IX(i, j - 1))) * N
    Next
Next

'Call set_bnd(1, velocX)
'Call set_bnd(2, velocY)

End Sub

Public Sub advect(b As Long, d() As Single, d0() As Single, velocX() As Single, velocY() As Single, dt As Single)

Dim i0 As Single
Dim i1 As Single
Dim j0 As Single
Dim j1 As Single

Dim dtx As Single
Dim dty As Single

Dim s0 As Single
Dim s1 As Single
Dim t0 As Single
Dim t1 As Single

Dim tmp1 As Single
Dim tmp2 As Single
Dim x As Single
Dim y As Single

Dim Nfloat As Single
Dim ifloat As Single
Dim jfloat As Single

Dim i As Long
Dim j As Long
Dim k As Long

Dim i0i As Long
Dim i1i As Long
Dim j0i As Long
Dim j1i As Long

dtx = dt * (N - 2)
dty = dt * (N - 2)

Nfloat = N

jfloat = 0
For j = 1 To N - 2
    jfloat = jfloat + 1
  
    ifloat = 0
    For i = 1 To N - 2
        ifloat = ifloat + 1
    
        tmp1 = dtx * velocX(IX(i, j))
        tmp2 = dty * velocY(IX(i, j))
        x = ifloat - tmp1
        y = jfloat - tmp2

        If x < 0.5 Then x = 0.5
        If x > Nfloat + 0.5 Then x = Nfloat + 0.5
        i0 = Fix(x)
        i1 = i0 + 1
    
        If y < 0.5 Then y = 0.5
        If y > Nfloat + 0.5 Then y = Nfloat + 0.5
        j0 = Fix(y)
        j1 = j0 + 1

        s1 = x - i0
        s0 = 1 - s1
        t1 = y - j0
        t0 = 1 - t1

        i0i = i0
        i1i = i1
        j0i = j0
        j1i = j1

        d(IX(i, j)) = s0 * (t0 * d0(IX(i0i, j0i)) + t1 * d0(IX(i0i, j1i))) + s1 * (t0 * d0(IX(i1i, j0i)) + t1 * d0(IX(i1i, j1i)))
    Next
Next

'Call set_bnd(b, d)
  
End Sub

Public Sub set_bnd(b As Long, x() As Single)

Dim i As Long
Dim j As Long

For i = 1 To N - 2
    x(IX(i, 0)) = IIf(b = 2, -x(IX(i, 1)), x(IX(i, 1)))
    x(IX(i, N - 1)) = IIf(b = 2, -x(IX(i, N - 2)), x(IX(i, N - 2)))
Next

For j = 1 To N - 2
    x(IX(0, j)) = IIf(b = 1, -x(IX(1, j)), x(IX(1, j)))
    x(IX(N - 1, j)) = IIf(b = 1, -x(IX(N - 2, j)), x(IX(N - 2, j)))
Next

x(IX(0, 0)) = 0.5 * (x(IX(1, 0)) + x(IX(0, 1)))
x(IX(0, N - 1)) = 0.5 * (x(IX(1, N - 1)) + x(IX(0, N - 2)))
x(IX(N - 1, 0)) = 0.5 * (x(IX(N - 2, 0)) + x(IX(N - 1, 1)))
x(IX(N - 1, N - 1)) = 0.5 * (x(IX(N - 2, N - 1)) + x(IX(N - 1, N - 2)))
  
End Sub

Public Sub Draw()

Dim SA As SAFEARRAY

Dim i As Long
Dim j As Long
Dim x As Long
Dim y As Long
Dim cx As Long
Dim cy As Long
Dim Angle As Single
Dim AngleX As Single
Dim AngleY As Single
Dim MyDens As Long

cx = N \ 2
cy = N \ 2

For i = -1 To 1
    For j = -1 To 1
        Call addDensity(cx + i, cy + j, 100 + Rnd * 50)
    Next
Next

For i = 0 To 1
    Angle = Rnd * 4 * PI
    AngleX = Cos(Angle)
    AngleY = Sin(Angle)
    AngleX = AngleX
    AngleY = AngleY
    Call addVelocity(cx, cy, AngleX, AngleY)
Next

Call step

With SA
    .cbElements = 1
    .cDims = 1
    .Bounds(0).lLbound = 0
    .Bounds(0).cElements = Bmp.bmWidthBytes * Bmp.bmHeight
    .pvData = Bmp.bmBits
End With

Call CopyMemory(ByVal VarPtrArray(ImageBuffer), VarPtr(SA), 4)

For i = 0 To UBound(ImageBuffer) Step 3
    MyDens = density(i \ 3)
    If MyDens > 255 Then MyDens = 255
    ImageBuffer(i) = MyDens
    ImageBuffer(i + 1) = MyDens
    ImageBuffer(i + 2) = MyDens
Next

Call CopyMemory(ByVal VarPtrArray(ImageBuffer), 0&, 4)

Form1.Picture1.Refresh

End Sub



















